package de.uni_marburg.powersort.MSort;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.stream.IntStream;

public class IMPL_M_2 {

    private IMPL_M_2() {
    }

    /**
     * Sorts the given range, using the given workspace array slice
     * for temp storage when possible. This method is designed to be
     * invoked from public methods (in class Arrays) after performing
     * any necessary array bounds checks and expanding parameters into
     * the required forms.
     *
     * @param a        the array to be sorted
     * @param lo       the index of the first element, inclusive, to be sorted
     * @param hi       the index of the last element, exclusive, to be sorted
     * @param c        the comparator to use
     * @param work     a workspace array (slice)
     * @param workBase origin of usable space in work array
     * @param workLen  usable size of work array
     * @since 1.8
     */
    protected static int MERGE_COST = 0;

    public static void fillWithAscRunsHighToLow(Integer[] A, int[] runLengths, int runLenFactor) {
        //A has a fixed size, but it doesn't have any meaningful values
        int n = A.length;
        //This ensures that the sum of runLengths multiplied by runLenFactor equals the list size n. If not, an AssertionError is thrown.
        assert IntStream.of(runLengths).sum() * runLenFactor == n;

        //System.out.println("IntStream Of run length output: "+IntStream.of(runLengths).sum());
        //IntStream.of(runLengths).forEach(System.out::println);
        for (int i = 0; i < n; i++) {
            //putting i in set a, while a is always the last index of n
            A[i] = n - i;
        }

        //For each value l in the array runLengths, do the following
        // runLengths = {2, 3, 5}, the loop will run three times, with l taking values 2, 3, and 5 respectively.
        int startIndex = 0;
        for (int l : runLengths) {
            int L = l * runLenFactor;
            // Sort the subarray from startIndex to startIndex+L manually
            for (int i = startIndex; i < startIndex + L - 1; i++) {
                for (int j = startIndex; j < startIndex + L - 1 - (i - startIndex); j++) {
                    if (A[j] > A[j + 1]) {
                        // Swap elements
                        int temp = A[j];
                        A[j] = A[j + 1];
                        A[j + 1] = temp;
                    }
                }
            }
            startIndex += L;
        }
    }

    static <T> void mergeWithGalloping(T[] a, int i, int m, int j, Comparator<? super T> c) {
        @SuppressWarnings("unchecked")
        T[] left = (T[]) new Object[m - i];
        @SuppressWarnings("unchecked")
        T[] right = (T[]) new Object[j - m];

        System.arraycopy(a, i, left, 0, m - i);
        System.arraycopy(a, m, right, 0, j - m);

        int k = i, l = 0, r = 0;
        int leftGallop = 1, rightGallop = 1;

        while (l < left.length && r < right.length) {
            if (c.compare(left[l], right[r]) <= 0) {
                a[k++] = left[l++];
                rightGallop = 1;
                leftGallop--;
            } else {
                a[k++] = right[r++];
                leftGallop = 1;
                rightGallop--;
            }

            // Enable galloping mode
            if (leftGallop <= 0 || rightGallop <= 0) {
                k += performGallopingMerge(a, left, right, k, l, r, c);
                leftGallop = rightGallop = 7;
            }
        }

        // Copy remaining elements
        while (l < left.length) a[k++] = left[l++];
        while (r < right.length) a[k++] = right[r++];
    }

    static <T> int performGallopingMerge(T[] a, T[] left, T[] right, int k, int l, int r, Comparator<? super T> c) {
        // Implement exponential search for faster merging
        int gallopAdded = 0;
        while (l < left.length && r < right.length) {
            int leftWins = exponentialBinarySearch(right, r, left[l], c);
            if (leftWins > 0) {
                System.arraycopy(left, l, a, k, leftWins);
                k += leftWins;
                l += leftWins;
                gallopAdded += leftWins;
            }

            if (l < left.length) {
                int rightWins = exponentialBinarySearch(left, l, right[r], c);
                if (rightWins > 0) {
                    System.arraycopy(right, r, a, k, rightWins);
                    k += rightWins;
                    r += rightWins;
                    gallopAdded += rightWins;
                }
            }
        }
        return gallopAdded;
    }

    static <T> int exponentialBinarySearch(T[] arr, int start, T key, Comparator<? super T> c) {
        // Implement exponential search with binary search for finding consecutive winning elements
        int index = start;
        int jump = 1;
        while (index + jump < arr.length && c.compare(arr[index + jump], key) <= 0) {
            index += jump;
            jump *= 2;
        }

        int right = Math.min(index + jump, arr.length);
        return binarySearchRange(arr, start, right, key, c);
    }

    static <T> int binarySearchRange(T[] arr, int left, int right, T key, Comparator<? super T> c) {
        while (left < right) {
            int mid = (left + right) >>> 1;
            if (c.compare(arr[mid], key) <= 0) {
                left = mid + 1;
            } else {
                right = mid;
            }
        }
        return left - left;
    }

    static <T> T[] merge(T[] run1, T[] run2, Comparator<? super T> c) {
        /**
         * We need this because:
         * In Java, you cannot create generic arrays directly (can't write new T[size])
         * The workaround is to create an Object array and cast it to T[]
         * This casting generates a warning because the compiler can't verify the type safety at runtime due to Java's type erasure
         * **/
        @SuppressWarnings("unchecked")
        T[] result = (T[]) new Object[run1.length + run2.length];
        int i = 0, j = 0, k = 0;

        while (i < run1.length && j < run2.length) {
            //This comparison only works if the lists are sorted
            if (c.compare(run1[i], run2[j]) <= 0) {
                result[k++] = run1[i++];
            } else {
                result[k++] = run2[j++];
            }
        }
        // can be improved by finding out which one is empty and only add the other one
        // Copy remaining elements
        while (i < run1.length) {
            result[k++] = run1[i++];
        }
        while (j < run2.length) {
            result[k++] = run2[j++];
        }

        return result;
    }

    // New helper method to handle range reversal
    private static <T> void reverseRange(T[] a, int start, int end) {
        // end is exclusive, so we need to use end-1 as the right boundary
        int left = start;
        int right = end - 1;
        while (left < right) {
            T temp = a[left];
            a[left] = a[right];
            a[right] = temp;
            left++;
            right--;
        }
    }

    static <T> void mergeInplace(T[] a, int i, int m, int j, Comparator<? super T> c) {
        // Minimize memory allocation and copying
        if (j - i <= 16) {
            insertionSortRange(a, i, j, c);
            return;
        }

        // Reduce temporary array allocations
        T[] merged = merge(Arrays.copyOfRange(a, i, m),
                Arrays.copyOfRange(a, m, j), c);
        System.arraycopy(merged, 0, a, i, merged.length);
    }


    static <T> int extendRun(T[] a, int i, Comparator<? super T> c) {
        // Faster, more aggressive run detection
        int j = i + 1;
        while (j < a.length && c.compare(a[j-1], a[j]) <= 0) {
            j++;
        }

        // Quick reversal for descending runs
        if (j < a.length && c.compare(a[j-1], a[j]) > 0) {
            reverseRange(a, i, j);
        }
        return j;
    }


    static <T> int determineRunType(T[] a, int start, Comparator<? super T> c) {
        if (start >= a.length - 1) return 0;
        int result = c.compare(a[start], a[start + 1]);
        return Integer.signum(result);
    }

    static <T> int findRunEnd(T[] a, int start, int runType, Comparator<? super T> c) {
        int j = start + 1;
        while (j < a.length) {
            int comparison = c.compare(a[j - 1], a[j]);
            if (runType == 0 && comparison > 0) break;
            if (runType < 0 && comparison > 0) break;
            if (runType > 0 && comparison < 0) break;
            j++;
        }
        return j;
    }

    public static int power(int[] run1, int[] run2, int n) {
        int i1 = run1[0], n1 = run1[1];
        int i2 = run2[0], n2 = run2[1];

        assert i1 >= 0;
        assert i2 == i1 + n1;
        assert n1 >= 1 && n2 >= 1;
        assert i2 + n2 <= n;

        double a = ((i1 + n1 / 2.0d) / n);
        double b = ((i2 + n2 / 2.0d) / n);

        int l = 0;
        while (Math.floor(a * Math.pow(2, l)) == Math.floor(b * Math.pow(2, l))) {
            l++;
        }
        return l;
    }


    static <T> void inPlaceMerge(T[] a, int start, int mid, int end, Comparator<? super T> c) {
        if (end - start <= 16) {
            insertionSortRange(a, start, end, c);
            return;
        }

        T[] temp = Arrays.copyOfRange(a, start, end);
        int i = 0, j = mid - start, k = start;

        while (i < mid - start && j < temp.length) {
            if (c.compare(temp[i], temp[j]) <= 0) {
                a[k++] = temp[i++];
            } else {
                a[k++] = temp[j++];
            }
        }

        while (i < mid - start) a[k++] = temp[i++];
        while (j < temp.length) a[k++] = temp[j++];
    }

    private static final int MIN_RUN_LENGTH = 32;


    public static <T> void powerSort(T[] a, Comparator<? super T> c) {
        int n = a.length;
        if (n < 64) {
            Arrays.sort(a, c);
            return;
        }

        List<int[]> runs = new ArrayList<>();
        int i = 0;
        int iterationLimit = n * 10;  // Increased limit
        int iterations = 0;

        while (i < n) {
            if (iterations++ > iterationLimit) {
                System.err.println("Iteration limit reached, falling back to standard sort");
                Arrays.sort(a, c);
                return;
            }

            int j = fastExtendRun(a, i, c);
            int p = calculateEffectivePower(runs, i, j - i, n);

            int mergeCycles = 0;
            while (runs.size() >= 2 && shouldMerge(runs, p)) {
                if (mergeCycles++ > 100) {
                    System.err.println("Merge cycle limit reached");
                    break;
                }
                mergeRuns(a, runs, c);
            }

            runs.add(new int[]{i, j - i, p});
            i = j;

            // Periodic debug output
            if (iterations % 100 == 0) {
                System.out.printf("Iteration %d: runs=%d, current index=%d%n",
                        iterations, runs.size(), i);
            }
        }

        while (runs.size() > 1) {
            mergeRuns(a, runs, c);
        }
    }

    private static boolean shouldMerge(List<int[]> runs, int newPower) {
        return runs.size() >= 2 &&
                runs.getLast()[2] >= newPower;
    }

    static <T> void mergeRuns(T[] a, List<int[]> runs, Comparator<? super T> c) {
        if (runs.size() < 2) return;

        int[] Y = runs.get(runs.size() - 2);
        int[] Z = runs.getLast();

        mergeInplaceFast(a, Y[0], Z[0], Z[0] + Z[1], c);

        runs.set(runs.size() - 2, new int[]{Y[0], Y[1] + Z[1], Y[2]});
        runs.removeLast();
    }

   

    private static boolean shouldAggressiveMerge(List<int[]> runs, int newPower) {
        return runs.size() >= 2 && runs.get(runs.size() - 1)[2] >= newPower;
    }

    static <T> void mergeOptimized(T[] a, List<int[]> runs, Comparator<? super T> c) {
        if (runs.size() < 2) return;

        int[] Y = runs.get(runs.size() - 2);
        int[] Z = runs.getLast();

        mergeInplaceFast(a, Y[0], Z[0], Z[0] + Z[1], c);

        runs.set(runs.size() - 2, new int[]{Y[0], Y[1] + Z[1], Y[2]});
        runs.removeLast();
    }









    private static <T> int fastExtendRun(T[] a, int start, Comparator<? super T> c) {
        int j = start + 1;
        while (j < a.length && c.compare(a[j-1], a[j]) <= 0) {
            j++;
        }

        if (j < a.length && c.compare(a[j-1], a[j]) > 0) {
            reverseRange(a, start, j);
        }
        return j;
    }

    private static int calculateEffectivePower(List<int[]> runs, int start, int length, int n) {
        if (runs.isEmpty()) return 0;

        int[] lastRun = runs.getLast();
        return (start + length/2) / (n/2);
    }


    static <T> void mergeInplaceFast(T[] a, int start, int mid, int end, Comparator<? super T> c) {
        if (end - start <= 16) {
            insertionSortRange(a, start, end, c);
            return;
        }

        T[] temp = Arrays.copyOfRange(a, start, end);
        int i = 0, j = mid - start, k = start;

        while (i < mid - start && j < temp.length) {
            if (c.compare(temp[i], temp[j]) <= 0) {
                a[k++] = temp[i++];
            } else {
                a[k++] = temp[j++];
            }
        }

        while (i < mid - start) a[k++] = temp[i++];
        while (j < temp.length) a[k++] = temp[j++];
    }

    static <T> void insertionSortRange(T[] a, int start, int end, Comparator<? super T> c) {
        for (int i = start + 1; i < end; i++) {
            T key = a[i];
            int j = i - 1;
            while (j >= start && c.compare(a[j], key) > 0) {
                a[j + 1] = a[j];
                j--;
            }
            a[j + 1] = key;
        }
    }
















    private static int calculateOptimizedPower(List<int[]> runs, int start, int length, int n) {
        if (runs.isEmpty()) return 0;

        int[] lastRun = runs.getLast();
        int l = 0;
        double a = (start + length / 2.0) / n;
        double b = (lastRun[0] + lastRun[1] / 2.0) / n;

        while (Math.floor(a * Math.pow(2, l)) == Math.floor(b * Math.pow(2, l))) {
            l++;
        }
        return l;
    }


    static <T> void mergeTopmost2(T[] a, List<int[]> runs, Comparator<? super T> c) {
        int[] Y = runs.get(runs.size() - 2);
        int[] Z = runs.getLast();

        mergeInplace(a, Y[0], Z[0], Z[0] + Z[1], c);

        runs.set(runs.size() - 2, new int[]{Y[0], Y[1] + Z[1], Y[2]});
        runs.removeLast();
    }
    public static int extendRunIncreasingOnly(List<Integer> a, int i) {
        if (i == a.size() - 1) {
            return i + 1;
        }
        int j = i + 1;
        while (j < a.size() && a.get(j - 1) <= a.get(j)) {
            j++;
        }
        return j;
    }

    public static int powerFast(int[] run1, int[] run2, int n) {
        // Bitwise optimizations
        int a = (run1[0] << 1) + run1[1];
        int b = a + run1[1] + run2[1];

        int l = 0;
        while (a < n) {
            a <<= 1;
            b <<= 1;
            l++;
            if (b >= n) break;
        }
        return l;
    }
}