package de.uni_marburg.powersort.MSort;

import java.util.Arrays;
import java.util.Comparator;

public class IMPL_M_3 {

    private static final int MIN_MERGE = 32;
    private static final int MIN_GALLOP = 7;

    private IMPL_M_3() {
    }

    public static void fillWithAscRunsHighToLow(Integer[] A, int[] runLengths, int runLenFactor) {
        int n = A.length;
        assert Arrays.stream(runLengths).sum() * runLenFactor == n;

        for (int i = 0; i < n; i++) {
            A[i] = n - i;
        }

        int startIndex = 0;
        for (int l : runLengths) {
            int L = l * runLenFactor;
            Arrays.sort(A, startIndex, startIndex + L);
            startIndex += L;
        }
    }

    private static <T> int extendRun(T[] a, int i, Comparator<? super T> c) {
        if (i >= a.length - 1) {
            return a.length; // Return the end of the array
        }

        int j = i + 1;
        boolean ascending = c.compare(a[i], a[j]) <= 0;

        while (j < a.length && c.compare(a[j - 1], a[j]) == (ascending ? -1 : 1)) {
            j++;
        }

        if (!ascending) {
            reverseRange(a, i, j);
        }

        return j;
    }

    private static <T> void reverseRange(T[] a, int start, int end) {
        end--;
        while (start < end) {
            T temp = a[start];
            a[start++] = a[end];
            a[end--] = temp;
        }
    }

    private static <T> void mergeInplace(T[] a, int i, int m, int j, Comparator<? super T> c, T[] temp) {
        int leftSize = m - i;
        int rightSize = j - m;

        // Validate indices
        if (leftSize < 0 || rightSize < 0) {
            throw new IllegalArgumentException("Invalid indices: leftSize=" + leftSize + ", rightSize=" + rightSize);
        }
        if (leftSize < 0) {
            throw new IllegalArgumentException("Invalid indices: leftSize is negative");
        }
        // Ensure the temporary array is large enough
        if (temp.length < leftSize) {
            temp = Arrays.copyOf(temp, leftSize);
        }

        System.arraycopy(a, i, temp, 0, leftSize);

        int li = 0, ri = m, k = i;
        int gallopCount = 0;

        while (li < leftSize && ri < j) {
            if (c.compare(temp[li], a[ri]) <= 0) {
                a[k++] = temp[li++];
                gallopCount++;
            } else {
                a[k++] = a[ri++];
                gallopCount = 0;
            }

            if (gallopCount >= MIN_GALLOP) {
                gallopCount = 0;
                while (li < leftSize && ri < j) {
                    if (c.compare(temp[li], a[ri]) <= 0) {
                        a[k++] = temp[li++];
                    } else {
                        a[k++] = a[ri++];
                    }
                }
                break;
            }
        }

        while (li < leftSize) a[k++] = temp[li++];
        while (ri < j) a[k++] = a[ri++];
    }

    public static <T> void powerSort(T[] a, Comparator<? super T> c) {
        int n = a.length;
        if (n < MIN_MERGE) {
            Arrays.sort(a, c);
            return;
        }

        // Initialize temporary array with a reasonable size
        T[] temp = (T[]) new Object[Math.min(n, MIN_MERGE)];
        int[] runStack = new int[40];
        int stackSize = 0;

        int i = 0;
        while (i < n) {
            int j = extendRun(a, i, c);

            // Ensure j > i
            if (j <= i) {
                throw new IllegalStateException("Invalid run: j <= i, i=" + i + ", j=" + j);
            }

            int[] newRun = new int[]{i, j - i};

            // Validate new run
            if (newRun[0] >= newRun[1]) {
                throw new IllegalArgumentException("Invalid run: start index >= length, i=" + i + ", j=" + j);
            }

            i = j;

            if (stackSize > 0) {
                int[] prevRun = new int[]{runStack[stackSize - 2], runStack[stackSize - 1]};
                int p = power(prevRun, newRun, n);

                while (stackSize > 0 && p <= runStack[stackSize - 1]) {
                    mergeInplace(a, runStack[stackSize - 2], runStack[stackSize - 1], i, c, temp);
                    stackSize -= 2;
                }
            }

            runStack[stackSize++] = newRun[0];
            runStack[stackSize++] = newRun[1];
        }

        while (stackSize > 2) {
            mergeInplace(a, runStack[stackSize - 4], runStack[stackSize - 3], n, c, temp);
            stackSize -= 2;
        }
    }

    private static int power(int[] run1, int[] run2, int n) {
        int i1 = run1[0], n1 = run1[1];
        int i2 = run2[0], n2 = run2[1];

        int a = 2 * i1 + n1;
        int b = a + n1 + n2;

        int l = 0;
        while (true) {
            l++;
            if (a >= n) {
                a -= n;
                b -= n;
            } else if (b >= n) {
                break;
            }
            a <<= 1;
            b <<= 1;
        }
        return l;
    }


}